import time

import torch
import torch.nn as nn

from .generator import Generator
from .preprocessing import Preprocessing


class LSTMGenerator(Generator):
    """
        LSTM model for text generation.
        Parameters
        ----------
        datasetObj : Preprocessing class object
            An object created using the class Preprocessing.

        nLayers: int
            The number of LSTM layers.
        
        batchSize: int
            The number of training instances per iteration.

        embSize: int
            The size of the vector in which each token will be embedded.
        
        lstmSize: int
            The number of features in the hidden state.
        
        epochs: int
            The number of cycles over the training data.
        
        dropout: int, optional, default 0
            The percentage of units that will be dropped per layer.
        
        embeddingType: string, optional, default fasttext
            The type of embeddings that will be used for the tokens.
            It can be glove, fasttext or default. Default means the NN.Embedding will be used.
        
        predictionIteration: int, optional, default 10
            The number of iterations for the prediction loop unless the model generates the stop token.
        
        glovePath: string, optional, default None
            The path for the GloVE embeddings if embeddingType is set to 'glove'.
        
        fastTextParams: dict, optional, default {}
            The parameters will be used while fine-tuning fastText to create the embeddings.
            Emtpy dictionary means the default training parameters will be used.
        
        loadFastText: string, optional, default None
            A path to the pretrained fastText model file.
        
        saveFastText: string, optional, default None
            A path to save the fastText model after training.
        
        fineTuneEmbs: bool, optional,  default False
            Whether to fine-tune the pretrained embeddings or not.

        selectionParams: dict, optional, default {"sType": 'topk', 'k': 5}
            sType: The type of selection method to use during prediciton. It is either 'topk' or 'nucleus', 't' and 'n' can also be used.
            k: in case of topk, it represents the k value.
            probThreshold: When using Nucleus selection, this represnts the probability threshold p. 

    """
    def __init__(self,datasetObj, nLayers, batchSize, embSize, lstmSize,
                   epochs, dropout = 0, embeddingType ='fasttext',
                   predictionIteration = 10, glovePath = None, fastTextParams = {},
                   loadFastText = None, saveFastText = None, fineTuneEmbs = False, 
                   selectionParams = {"sType": 'topk', 'k': 5}):

        super(LSTMGenerator, self).__init__(datasetObj,nLayers,batchSize, embSize,
                                           lstmSize,epochs,dropout,embeddingType,predictionIteration,
                                           glovePath,fastTextParams,loadFastText,saveFastText,fineTuneEmbs,
                                           selectionParams)

        self.lstm_size = self.rnnSize
        self.predIter = predictionIteration

        self.lstm = nn.LSTM(self.embSize,
                        self.lstm_size,
                        num_layers = self.nLayers,
                        dropout = self.dropout,
                        batch_first = True)
        self.dense = nn.Linear(self.lstm_size, self.n_vocab)


    def forward(self, x, length, prev_state):
        """
            params:
                x: the tensor of input instances.
                length: the length that will be used from each training instance to handle paddings.
                prev_state: the hidden and the cell states

        """

        embed = self.embedding(x)
        packed_input = self.pack_src(embed, length)
        packed_output, state = self.lstm(packed_input, prev_state)
        padded,_ = self.pad_pack(packed_output) 
        logits = self.dense(padded)
        return logits, state

    def info(self):
        """
            Show the parameters used with the generator.
        """

        super(LSTMGenerator, self).info('LSTM')
    
    def run(self):
        """
            Training method
        """

        criterion, optimizer = self.get_loss_and_train_op(0.01)

        iteration = 0
        total_iterations = (len(self.db) // self.batchSize) * self.epochs

        for e in range(self.epochs):        
            batches = self.get_batches()
            
            state_h, state_c = self.zero_state(2, self.batchSize)
            state_h = state_h.to(self.device)
            state_c = state_c.to(self.device)
            
            batch_times = []
            for batch in batches:

                start_time = time.time()
                src_lengths, _ ,src_batch,trg_batch = self.get_src_trg(batch)
                
                iteration += 1

                self.train()


                optimizer.zero_grad()
                
                trg_batch = torch.t(trg_batch).to(self.device)
                src_batch = src_batch.to(self.device)
                
                logits, (state_h, state_c) = self(src_batch,src_lengths,(state_h,state_c))

                loss = criterion(logits.transpose(1, 2), trg_batch)

                state_h = state_h.detach()
                state_c = state_c.detach()

                loss_value = loss.item()

                loss.backward(retain_graph=True)

                optimizer.step()

                loss.backward()

                _ = torch.nn.utils.clip_grad_norm_(
                self.parameters(), 5)

                optimizer.step()

                batch_time = time.time() - start_time
                batch_times.append(batch_time)

                if iteration % 10 == 0:
                    mean_time = sum(batch_times)/len(batch_times)
                    remaining_iters = total_iterations - iteration       
                    remaining_seconds = remaining_iters * mean_time 
                    remaining_time = time.strftime("%H:%M:%S",
                        time.gmtime(remaining_seconds))
                    progress = "{:.2%}".format(iteration/total_iterations)
                    print('Epoch: {}/{}'.format(e+1, self.epochs),
                          'Progress:', progress,
                          'Loss: {}'.format(loss_value),
                          'ETA:', remaining_time)


    def generate_document(self, predIter = None, selection = None, k = None, prob = None):
        """
            Generate documents.
            selection: The type of selection will be used. its either topk or nucleus, t and n can also be used.
            k: In case of topk, it represents the k value.
            prob: In case of nucleus, its represnts the probability thresholds 
            predIter: The number of iterations for the prediction loop unless the model generated a stop token.


            They are optional while calling this method. The values passed while creating the generator object will be used.
        """
        return super(LSTMGenerator, self).generateDocument('LSTM', predIter, selection, k, prob)




import os
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.utils.rnn as rnn_utils
from random import random
import numpy as np

from .embedding import PretrainedEmbeddings
from .preprocessing import Preprocessing


class Generator(nn.Module):

    def __init__(self, datasetObj, nLayers, batchSize, embSize, rnnSize,
                   epochs , dropout , embeddingType ='fasttext', predictionIteration = 10,
                   glovePath = None, fastTextParams = {},
                   loadFastText = None, saveFastText = None, fineTuneEmbs = False, 
                   selectionParams = {"sType": 'topk', 'k': 5, 'probThreshold': 0.5}):
        super(Generator, self).__init__()

        self.db = self.__checkDataset(datasetObj)

        self.device =  torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        self.vocabToInt = self.db.DataVocab.vocab.stoi
        self.intToVocab = self.db.DataVocab.vocab.itos

        self.predIter = self.__checkIntParam(predictionIteration, 'predictionIteration')
        self.batchSize = self.__checkIntParam(batchSize,'batchSize')
        self.embSize = self.__checkIntParam(embSize, "embSize")
        self.nLayers = self.__checkIntParam(nLayers, "nLayers")
        self.epochs = self.__checkIntParam(epochs, "epochs")
        self.rnnSize = self.__checkIntParam(rnnSize, "lstm_size/gru_size")
        self.fineTuneEmbs = fineTuneEmbs


        self.glovePath = self.__checkPath(glovePath, "glovePath")
        self.loadFastText = self.__checkPath(loadFastText, "loadFastText")
        self.saveFastText = self.__checkPath(saveFastText, "saveFastText")

        self.dropout = self.__checkDropout(dropout)
        self.embeddingType = self.__checkEmbedding(embeddingType)
        self.fastTextParams = self.__checkDictParams(fastTextParams)

        self.n_vocab = len(self.db.DataVocab.vocab)
        self.embedding = self.__getEmbs()
        
        self.selection = None
        self.topk = 5
        self.nucluesProb = 0.5
        self.__checkSelection(selectionParams)

    def __checkSelection(self,value):
        self.__checkDictParams(value)
        if not len(value):
            raise Exception("The parameter cannot be empty")

        if 'sType' not in value:
            raise Exception("The dictionary has to have the key sType that indicates the selection type")
        
        else:
            selection = value['sType'].lower()
            if selection in 'topk':
                self.selection = 'topk'
                if 'of' in value:
                    self.topk = value['k']

            elif selection in 'nucleus':
                self.selection = 'nucleus'
                if 'probThreshold' in value:
                    self.nucluesProb = self.__checkProb(value['probThreshold'])
            else:
                raise Exception("sType can only be topk or nucleus")

    def __checkProb(self,value):
        if sum([type(value) == float, 0 <= value <= 1]) != 2:
            raise Exception("Make sure the probThreshold is a positive float, between 0 and 1." )
        return value

    def __checkDataset(self, dataset):
        filetype = type(dataset).__base__.__name__
        if filetype != "Dataset":
            raise Exception("Please provide a Dataset object. Instead, got '%s'."% filetype)
        return dataset
    
    def __checkEmbedding(self,value):
        value = value.lower()
        if value in 'glove': # both glove and g will be accepted.
            return 'glove'
        elif value in 'fasttext':
            return 'fasttext'
        elif value in 'random':
            return 'random'

        raise Exception("Embedding type can only be glove, fasttext or random(nn.Embedding)")

    def __getEmbs(self):
        if self.embeddingType in ['glove', 'fasttext']:
            pretrainedEmbeddings = PretrainedEmbeddings(self.db,self.embSize ,self.embeddingType,
                self.glovePath, self.fastTextParams, self.loadFastText, self.saveFastText)
            
            weights = pretrainedEmbeddings.weights
            
            self.__checkembedDims(weights.shape[1])
                        
            return self.create_emb_layer(weights, non_trainable=self.fineTuneEmbs)

        else:
            return nn.Embedding(self.n_vocab, self.embSize, padding_idx=0)
    
    def __checkembedDims(self, value):
        if self.embSize != value:
            raise Exception("The pretrained embedding size does not equal the input size that the rnn will get\n"
                            f"pretrained embedding size = {value} , LSTM input = {self.embSize}")
        return value

    def create_emb_layer(self, weights_matrix, non_trainable=False):
        num_embeddings, embedding_dim = weights_matrix.shape
        emb_layer = nn.Embedding(num_embeddings, embedding_dim,padding_idx=0)
        emb_layer.weight.data.copy_(torch.from_numpy(weights_matrix))
        if non_trainable:
            emb_layer.weight.requires_grad = False

        return emb_layer
    
    def __checkIntParam(self, param, name):
        if sum([type(param) != int, param <= 0]) >= 1:
            raise Exception("Make sure %s is a positive int."%name)
        return param

    def __checkDropout(self, dropout):
        if self.nLayers < 2 and dropout != 0:
            raise Exception("The network must have more than 1 layer to have dropout.")
        if dropout < 0 or dropout > 1:
            raise Exception("Dropout value must be between 0 and 1.")     
        return dropout   

    def __checkDictParams(self, value):
        if type(value) != dict:
            raise Exception("Please provide a proper dict containing the parameters.")
        return value
    
    def __checkPath(self, param, name):
        if not param : return None
        elif type(param) != str:
            raise Exception("Please provide a proper path for %s"%name)
        return param

    def get_batches(self):
        num_batches = len(self.db) // self.batchSize
        for i in range(0, num_batches * self.batchSize, self.batchSize):
            yield self.db[i:i+self.batchSize]

    def pack_src(self, embed,length):
        return rnn_utils.pack_padded_sequence(embed, length, batch_first=True)
    
    def pad_pack(self, packed_output):
        return rnn_utils.pad_packed_sequence(packed_output)
    
    def pad_seq(self, batch):
        return torch.nn.utils.rnn.pad_sequence(batch, batch_first=True).long()

    def get_loss_and_train_op(self, lr=0.001):
        criterion = nn.CrossEntropyLoss()
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)

        return criterion, optimizer

    def get_src_trg(self, batch):
        """
         batches shape: batch_size * tokens
        """        
        src_batch,src_length = self.prepare_batch(batch)
        trg_batch,trg_length = self.prepare_batch(batch, IsSrc=False)
        
        return src_length, trg_length, src_batch, trg_batch

    def prepare_batch(self, example_objects,IsSrc=True):
        """
         turn the tokens into padded seqs and get the length of each seq separately.
        """
        target_token,target_length = ('src_token','src_length') if IsSrc else ('trg_token','trg_length')
            
        batch = self.pad_seq([torch.Tensor(getattr(exp,target_token)) for exp in example_objects])
        
        lengths = torch.tensor([getattr(exp,target_length) for exp in example_objects])
        
        lengths , perm_idx = lengths.sort(0, descending=True)
        
        batch = batch[perm_idx]
        
        return batch, lengths

    def zero_state(self, n, batchSize):
        return ([torch.zeros(self.nLayers, batchSize, self.rnnSize)]) * n      

    def select_nucleus(self, out, p = None):
        p = (p or self.nucluesProb)
        probs = F.softmax(out, dim=-1)[0]
        idxs = torch.argsort(probs, descending = True)
        res,prob, cumsum = [], [], 0.
        for idx in idxs:
            res.append(idx)
            prob.append(probs[idx].item())
            cumsum+=probs[idx]
            if cumsum > p:
                break
        prob = prob / np.sum(prob)
        choice = np.random.choice(res, 1, p = prob)[0]
        return choice

    def choose_from_top(self, probs, n=5):
        probs = F.softmax(probs, dim=-1)[0]
        ind = np.argpartition(probs, -n)[-n:]
        top_prob = probs[ind]
        top_prob = top_prob / torch.sum(top_prob)
        choice = np.random.choice(n, 1, p = top_prob.numpy())
        token_id = ind[choice][0]
        return int(token_id)
    
    def info(self, generatorType):
        print(f"Size of data: {len(self.db)} documents.")
        print("Device used:", self.device)
        print("Vocabulary size:", self.n_vocab)
        print(f"Layers of {generatorType}:", self.nLayers)
        print("Dropout:", f"True, {self.dropout}" if self.dropout>0 else False)
        print(f"{generatorType} size:", self.rnnSize)
        print("Batch size:", self.batchSize)
        print("Embedding type:", self.embeddingType)
        print("Fine tune embeddings?:", self.fineTuneEmbs)
        print("Embedding size:", self.embSize)
        print("Selection method:", f"{self.selection}, k = {self.topk}" if self.selection == "topk"
            else f"{self.selection} p = {self.nucluesProb}")


    def save(self, path):
        directory = "/".join(path.split('/')[:-1])
        if not os.path.exists(directory):
            os.mkdir(directory)

        torch.save({
            'nLayers': self.nLayers,
            'embSize': self.embSize,
            'rnnSize': self.rnnSize,
            'dropout': self.dropout,
            'model_state_dict': self.state_dict(),
            }, path)

    def load(self, path):
        saved_model = torch.load(path)
        model_state_dict = saved_model["model_state_dict"]
        nLayers = saved_model["nLayers"]
        embSize = saved_model["embSize"]
        rnnSize = saved_model["rnnSize"]
        if sum([self.nLayers!=nLayers, 
                self.embSize!=embSize,
                self.rnnSize!=rnnSize]) > 0:
            raise Exception("Please create a model with the same parameters as the trained one. Trained model parameters: {} nLayers, {} embSize, {} lstmSize/gruSize".format(nLayers,embSize,rnnSize))

        self.load_state_dict(model_state_dict)

    
    def generateDocument(self, modelName, predIter = None, selection = None, k = None, prob = None):
        selection = (selection or self.selection)
        k = (k or self.topk)
        prob = (prob or self.nucluesProb)
        predIter = (predIter or self.predIter)
        eos =  self.vocabToInt[self.db.DataVocab.eos_token]


        self.eval()

        words = self.db.getSeed()
        choice = self.vocabToInt[words[0]]

        if modelName == "GRU":
            states = self.zero_state(1, 1)[0]        
        else:
            states = self.zero_state(2, 1)
        with torch.no_grad():
            for _ in range(predIter):
                ix = torch.tensor([[choice]]).long().to(self.device)
                length = torch.tensor([ix.shape[1]])    

                output, states = self(ix,length,states)

                if selection in 'nucleus':
                    choice = self.select_nucleus(output.view(1,-1),p=prob)

                elif selection in 'topk':
                    choice = self.choose_from_top(output.view(1,-1),n=k)
                    
                if choice == eos:
                    words.append(self.intToVocab[eos])
                    return ' '.join(words)
                        
                words.append(self.intToVocab[choice])

            return ' '.join(words)



        





    



import os

import numpy as np

from .importers import importFasttext


class PretrainedEmbeddings:
    """
        Get pretrained embeddings using GloVe or fastText.
        
        Parameters
        ----------
        datasetObj : Preprocessing class object
            An object created using the class Preprocessing.
        
        embSize: int
            The size of the vector in which each will be embedded.

        glovePath: string, optional, default None
            The path for the glove embeddings if embeddingType is set to 'glove'.
        
        fastTextParams: dict, optional, default {}
            The parameters will be used while fine-tuning fastText to create the embeddings.
            Emtpy dictionary means the default training parameters will be used.
        
        loadFastText: string, optional, default None
            A path to the pretrained fastText model file.
        
        saveFastText: string, optional, default None
            A directory path to save the model with the name fastText_{dim}.bin
            or a file path that will have a filename with the extension .bin
    """

    def __init__(self,datasetObj, embSize = 64,
                                 embeddingType = 'fasttext',
                                 glovePath = None,
                                 fastTextParams = {},
                                 loadFastText = None,
                                 saveFastText = None,
                                  ):
        

        self.__preprocessingObj(datasetObj)
        self.__Datavocab = datasetObj.DataVocab.vocab.stoi
        self.__dataFile = datasetObj.fileName
        self.embSize = embSize


        self.__isGlove = self.__checkEmbedding(embeddingType)

        # fastText
        self.__loadFastText = self.__checkModelFile(loadFastText)
        self.__saveFastText = self.__saveFastTextParam(saveFastText)
        self.__fastTextParams = fastTextParams
        self.__fastTextModel = None

        # Glove
        self.__gloveEmb = {}
        self.glovePath = glovePath


        self.weights = self.__getEmbs()
    
    def __getEmbs(self):
        if self.__isGlove:
            self.__checkGloveParams()
            self.__getGlove()
            
        else:
            self.__getFast()

        return self.__getMatrix()

    def __checkEmbedding(self,value):
        value = value.lower()
        if value in 'glove': # both glove and g will be accepted.
            return True
        elif value in 'fasttext':
            self.__fasttext = importFasttext()
            return False

        raise Exception("Embedding type can only be glove or fasttext")

    def __preprocessingObj(self,preprobObj):
        if preprobObj.__class__.__name__ != 'Preprocessing':
            raise Exception("Please provide the dataset object of type Preprocessing.")
        return preprobObj

    def __checkModelFile(self,fileName):
        """
            ensures that the extension of fasttext trained file is .bin
        """
        if type(fileName) == str:
            if fileName.split(".")[-1] != 'bin':
                raise Exception("FastText trained model should have a .bin extension")
            return fileName
        return None


    def __embeddingType(self):
        return 'Glove' if self.__isGlove else 'FastText'

    def __getFast(self):
        
        if self.__loadFastText:
            self.__fastTextModel = self.__fasttext.load_model(self.__loadFastText)
            print("Loaded the model located at '{m}'".format(m= self.__loadFastText))

        else:
            self.__trainFast()

    def __saveFastTextParam(self, value):
        if value:
            if os.path.isdir(value):
                modelName = "fastText_{dim}.bin".format(dim=self.embSize)
                return os.path.join(value,modelName)
            elif value.endswith('.bin'):
                return value
            else:
                raise Exception("SaveFastText param is invalid, its neither a dirc path nor a .bin file path")
        return None

    def __trainFast(self):
        print("{pre_emb}: Training model".format(pre_emb=self.__embeddingType()))
        self.__fastTextModel = self.__fasttext.train_unsupervised(input=self.__dataFile,dim=self.embSize,**self.__fastTextParams)

        if self.__saveFastText:
            self.__fastTextModel.save_model(self.__saveFastText)
    

    def __checkGloveParams(self):
        """
            if isGlove is true, then glove path has to be provided
        """
        if self.__isGlove and not self.glovePath:
            raise Exception("Glove file path is needed when setting embeddingType=glove.")
    

    def __getGlove(self):
        with open(self.glovePath, 'r',encoding="utf8") as f:
            for line in f:
                values = line.split()
                word = values[0]
                if word in self.__Datavocab:
                    self.__gloveEmb[word] =  np.asarray(values[1:], "float32")
            
            if len(values[1:]) != self.embSize:
                self.embSize = len(values[1:])
                print("Setting embedding size using Glove file.")
                    
        print("{pre_emb}: READ {t} tokens".format(pre_emb=self.__embeddingType(),t = len(self.__gloveEmb.keys())))
    


    def __getMatrix(self):
        counter = 0
        weights_matrix = np.zeros((len(self.__Datavocab),self.embSize))    
        for index, word in zip(self.__Datavocab.values(),self.__Datavocab):

            if self.__isGlove:
                if word not in self.__gloveEmb:
                    weights_matrix[index] = np.random.normal(scale=0.6, size=(self.embSize, ))
                else:
                    weights_matrix[index] = self.__gloveEmb[word]
                    counter += 1
            else:
                weights_matrix[index] = self.__fastTextModel.get_word_vector(word)
                counter += 1
        
        weights_matrix[:2] = np.zeros(self.embSize,)
        print("{pre_emb}: FOUND {c} tokens out of {t}".format(pre_emb=self.__embeddingType(),c=counter,t=len(self.__Datavocab)))
        return weights_matrix



def importFasttext():
    try:
        import fasttext as ft
        print("Import fasttext")
    except ImportError:
        ft = None
    
    if not ft:
        raise ImportError("Couldn't import fasttext.")

    return ft

def importNltk():
    try:
        from nltk.tokenize import word_tokenize as wt
        print("import nltk.tokenize.word_tokenize function")
    except ImportError:
        wt = None

    if not wt:
        raise ImportError("Couldn't import nltk.tokenize.word_tokenize function.")
    
    return wt




import csv
import json
import os
import re
from collections import Counter

import numpy as np
from torchtext.data import Dataset, Example, Field

from .importers import importNltk


class DocumentExample(Example):
    def __init__(self,src,trg):
        self.src = src
        self.trg = trg
        self.src_token = None
        self.trg_token = None
        self.src_length = len(self.src)
        self.trg_length = len(self.trg)
    
    def create_tokens(self,field):        
        self.src_token = [field.vocab.stoi[word] for word in self.src]
        self.trg_token = [field.vocab.stoi[word] for word in self.trg] + [0]  


class Preprocessing(Dataset):
    """
        Preprocessing the documents, creating paris of src and trg and building the vocab.

        Parameters
        ----------
        fileName: string
            The path to a text file with training documents/instances separated by newline (\n).

        spacyObj: spaCy language class object, optional, default None
            The spacy language class object that will be used in recognizing entities
            and grouping them into one token. 

        instanceMxLen: int, optional , default None
            Used to ignore documents that are larger than a certain length of tokens.
        
        fieldParams: dict, optional, default {'lower': True}
            The parameters that will be used in creating the Field object.
        
        seedParams: dict, optional, default {'N_first': 1, 'minFreq': 5}
            Parameters that will be used while selecting the random seed used to initalize prediction.
                N_first: get the first n_first tokens of each training document.
                minFreq: The min frequency a seed must have before it is included in the pool. 
            
            by default random seed is enabled. If seedParams = {} then static seed will be applied.

        txtSeparator: str, optional, default '\n'   
            When the data file is a txt file, this is the character serparating documents.
            By default, the assumption is that each document is on a separate line.

        csvIndex: int, optional, default None
            When the data file is a csv file, this is the index of the column to parse from.

        jsonKey: str, optional, default None
            When the data file is a csv file, this is the json key to parse from.
            Typically, it is 'body' or 'text'.

    """

    __tokPattern = r"""[0-9A-Za-z_]*[A-Za-z_-]+[0-9A-Za-z_]*|\.|\!|\?|\d+|\-|%|[.,!?;'"]"""

    def __init__(self, fileName,
                    tokenizationMethod = 'regex',
                    spacyObj = None,
                    instanceMxLen = None,
                    fieldParams = {'lower': True, 'eos_token': '<!EOS!>'}, 
                    seedParams = {'N_first': 1, 'minFreq': 5},
                    txtSeparator = '\n',
                    csvIndex = None,
                    jsonKey = None):


        # To be set by the user
        self.csvIndex = csvIndex
        self.jsonKey = jsonKey
        self.txtSeparator = txtSeparator

        self.fileName = self.__checkFileName(fileName)
        self.instanceMxLen = instanceMxLen
        self.seedParams = self.__checkSeedParams(seedParams)


        # class variables
        self.fieldParams = fieldParams
        self.examples = None
        self.__nlp = self.__checkSpacyObj(spacyObj)
        self.__DataVocab = Field(**self.fieldParams)
        self.__text = None


        self.__customTokenize = self.__tokenizationMethod(tokenizationMethod)

        self.__readFile()


    def info(self):
        print("************INFO*************")
        if self.extension == "json":
            print(f"File name: {self.fileName}, jsonKey = {self.jsonKey}")
        elif self.extension == "txt":
            print(f"File name: {self.fileName}, txtSeparator = {repr(self.txtSeparator)}")
        else:
            print(f"File name: {self.fileName}, csvIndex = {self.csvIndex}")
        print("Tokenization method:", self.__customTokenize.__name__)
        print("Ignore long documents:", f"True, where length > {self.instanceMxLen}" if self.instanceMxLen!= None 
                                                    else False)
        print("Seed parameters:", self.seedParams)
        print("Field parameters:", self.fieldParams)

    
    @property
    def DataVocab(self):
        return self.__DataVocab

    def __checkFileName(self, fileName):
        extension = fileName.split(".")[-1]
        
        if extension not in ['txt', 'csv', 'json']:
            raise Exception("Only works with txt, csv, and json files.")

        if extension == 'csv' and self.csvIndex == None:
            raise Exception("Please provide a 'csvIndex' to indicate the column to parse from.")
        elif extension == 'json' and self.jsonKey == None:
            raise Exception("Please provide a 'jsonKey' to indicate json key to parse from.")
        
        self.extension = extension
        return fileName
    
    def __checkSeedParams(self,params):
        attr_list = ['N_first', 'minFreq']

        if not len(params):
            return {}
        
        elif not all(attr in params for attr in attr_list):
            raise Exception("seedParams should contain the attributes {attrs}".format(attrs = attr_list))

        return params
    
    def __checkSpacyObj(self,spacyObj):

        if spacyObj:
            if type(spacyObj).__base__.__name__ != 'Language':
                raise Exception("spacyObj has to be of a type Language")

            return spacyObj

        return None

    def __len__(self):
        return len(self.examples)
    
    def __getitem__(self, id):
        return self.examples[id]

    def __spacyTokenization(self,instance):

        return [entity.text.strip() for entity in self.__nlp(instance) if entity.text.strip() ]


    def nltkTokenization(self, document):
        return self.word_tokenize(document)

    def regexTokenization(self, document):
        return re.findall(self.__tokPattern, document)

    def __tokenizationMethod(self,value):
        value = value.lower()
        if value in 'nltk':
            self.word_tokenize = importNltk()
            return self.nltkTokenization
        elif value in 'regex':
            return self.regexTokenization
        elif value in 'spacy':
            if not self.__nlp:
                raise Exception("The spacy object is not provided to tokenize using spacy.")
            return self.__spacyTokenization
        
        raise Exception("The param 'tokenizationMethod' can only be nltk, regex and spacy")

    def __tokenize(self,instance):

        instance = self.__customTokenize(instance)

        return {'src': instance, 'trg': instance[1: ]}

    def __lengthLimit(self, instance):
        length = len(instance.split()) 
        if self.instanceMxLen:
            return length <= self.instanceMxLen and length > 0
        return length > 0


    def __getObjects(self, rawDocuments):        
        self.fields = {"src": self.__DataVocab}
        return [DocumentExample(**self.__tokenize(instance)) 
                for instance in rawDocuments
                if self.__lengthLimit(instance)]

    def __getStartWords(self): 
        n_first = 1 if 'N_first' not in self.seedParams else self.seedParams['N_first']

        start_words = [' '.join(instance.split()[:n_first])
                           for instance in self.__text
                           if len(instance.split()) > n_first]
        
        freq_start = Counter(start_words)
        if not len(self.seedParams):
            word, _ = freq_start.most_common(1)[0]
            print("The word {w} is the Static Seed. ".format(w=word))
            return {word: 1.0}
        
        return {word:freq/len(self.examples) for word,freq  in freq_start.items() if freq >= self.seedParams['minFreq'] }

    def __readFile(self):
        file_obj = open(self.fileName)

        if self.extension == "txt":
            self.__text = file_obj.read().split(self.txtSeparator)
        elif self.extension == "csv":
            reader = csv.reader(file_obj)
            self.__text = [row[self.csvIndex] for row in reader]
        else:
            json_data = json.loads(file_obj.read())
            self.__text = [row[self.jsonKey] for row in json_data]

        self.examples = self.__getObjects(self.__text)
        self.seeds =  self.__getStartWords()
        self.__build_vocab()

    def __prepare_tokens(self):
        for instance in self.examples:
            instance.create_tokens(self.__DataVocab) 

    def __build_vocab(self):
        self.__DataVocab.build_vocab(self)
        self.__DataVocab.vocab.stoi['<unk>'] = 1
        self.__DataVocab.vocab.stoi['<pad>'] = 0
        self.__DataVocab.vocab.itos[0] = '<pad>'
        self.__DataVocab.vocab.itos[1] = '<unk>'
        self.__prepare_tokens() 

    def getSeed(self):
        """
            return a weighted seed. 
            In case static seed is enabled, then the most frequent token will be the seed.
        """
        seeds = list(self.seeds.keys())
        probs = list(self.seeds.values())
        return np.random.choice(seeds, 1 , probs).tolist()
